/*
 *  dice - DESCRIPTION OF PACKAGE
 *  Copyright (C) 2003 RossLodge
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  The author of this program grants you the ability to use this code
 *  in conjunction with code that is covered under the Open Gaming License
 *
 *  Roll.java
 *
 *  Created on Oct 18, 2003, 1:34:56 PM
 */
package com.electronicmuse.djep.function;
 
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import com.electronicmuse.djep.*;
import com.electronicmuse.djep.ParseException;

import edu.cornell.lassp.houle.RngPack.RandomElement;


/**
 * @author RossLodge
 * 
 * <p>This class forms the basis for the dJEP extensions to the JEP library.
 * It evaluates a <code>ROLL</code> token, which is an operator that comes
 * in precedence between multiplicative and additive operators.</p>
 * 
 * <p>The class receives two parameters, the number of rolls and the number of
 * faces per die, as in "3d6," or whatever.  It initializes a random number
 * generator, which must be a subclass of <code>edu.cornell.lassp.houle.RngPack.RandomElement</code>.
 * A default randomizer class is provided that wraps the java.util.Random class.
 * The class used may be changed via setRandomClassName.</p>
 * 
 * <p>The class provides very minimal retrieval of individual rolls, via the
 * <code>getRolls()</code> method.</p>
 */
public class Roll extends PostfixMathCommand
{
	
	/**
	 *
	 * Holds the result values for a call to the Roll() function.
	 * This class has two fields, the number of rolls and the faces on the dice.
	 * It also holds a list of the original results of each roll.
	 *
	 * @author Ross Lodge
	 */
	public class RollResult extends Object
	{
		/**
		 * A linked list of rolls results.
		 */
		private LinkedList m_rolls = new LinkedList();
		/**
		 * Number of rolls (i.e., the 3 in "3d6").
		 */
		private int m_numberOfRolls = 0;
		/**
		 * Faces on the dice (i.e., the 6 in "3d6").
		 */
		private int m_faces = 0;
		
		/**
		 * Constructs the object.
		 * 
		 * @param rolls Number of rolls
		 * @param faces Faces on the dice
		 */
		public RollResult(int rolls, int faces)
		{
			m_numberOfRolls = rolls;
			m_faces = faces;
		}

		/**
		 * @return The faces on the dice
		 */
		public int getFaces()
		{
			return m_faces;
		}

		/**
		 * @return The number of rolls
		 */
		public int getNumberOfRolls()
		{
			return m_numberOfRolls;
		}

		/**
		 * @return A list, guaranteed non-null, which contains
		 * Integer objects representing the results of each roll.
		 */
		public List getRolls()
		{
			return m_rolls;
		}

		/**
		 * Adds a new roll to the end of the rolls list.
		 * 
		 * @param i int representing the result of a roll.
		 * @return Success/failure of adding the integer to the list
		 */
		public boolean add(int i)
		{
			return m_rolls.add(new Integer(i));
		}
		
		public String toString()
		{
			StringBuffer returnValue = new StringBuffer(m_numberOfRolls + "d" + m_faces + ": ");
			for (Iterator i = m_rolls.iterator(); i.hasNext();)
			{
				if (returnValue.toString().endsWith(": "))
				{
					returnValue.append(i.next().toString());
				}
				else
				{
					returnValue.append(", " + i.next().toString());
				}
			}
			return returnValue.toString();
		}

	}
	
	/** Constant for the constructure we're looking for. */
	private static final Class[] DESIRED_PARAMETER_ARRAY = {Date.class};
	/** Whether or not we are saving rolls to the roll list. */
	private static boolean m_logging = false;
	/** Default class name to instantiate. */
	private static String m_randomClassName = "edu.cornell.lassp.houle.RngPack.Ranlux";
	/** A list of rolls for the logging facility */
	private static LinkedList m_rolls = new LinkedList();
	/**
	 * The current randomizer for the class.  Static because the randomizers are
	 * initialized with time, and if they are initialized very close to one another
	 * they might generate the same number. */
	private static RandomElement m_random = null;

	/**
	 * <p>Returns the current random class name.</p>
	 * 
	 * @return Class name of randomizer class
	 */
	public static String getRandomClassName()
	{
		return m_randomClassName;
	}

	/**
	 * <p>Returns a list of the RollResult values for rolls, and clears
	 * the internal list.</p>
	 * 
	 * @return A list of the most recent rolls, from oldest to newest
	 */
	public static List getRolls()
	{
		List returnValue = m_rolls;
		m_rolls = new LinkedList();
		return returnValue;
	}

	/**
	 * <p>Returns whether or not the class is currently logging rolls
	 * to the static list.</p>
	 * 
	 * @return State of logging facility.
	 */
	public static boolean isLogging()
	{
		return m_logging;
	}

	/**
	 * <p>Turns logging on or off.</p>
	 * 
	 * @param b <code>true</code> to turn logging on; <code>false</code> to turn
	 * it off.
	 */
	public static void setLogging(boolean b)
	{
		m_logging = b;
	}

	/**
	 * <p>Sets the random class name.</p>
	 * 
	 * @param string The fully qualified class name of a class derived
	 * from <code>edu.cornell.lassp.houle.RngPack.RandomElement</code>.
	 */
	public static void setRandomClassName(String string)
	{
		m_randomClassName = string;
	}
	
	/**
	 * 
	 * <p>The default (and only) constructor.  Sets the number
	 * of parameters for the JEP package, and calls <code>setRandom()</code> to
	 * initialize the randomizer.</p>
	 */
	public Roll()
	{
		numberOfParameters = 2;
		setRandom();
	}

	/**
	 * <p>The run command for the JEP framework.  This receives the arguments
	 * to the operator as a stack, pops off the two it needs, does type checking,
	 * rolls the dice on the randomizer, and pushes the result back onto the stack.
	 * Logging is performed if it is turned on.</p>
	 * 
	 * @see org.nfunk.jep.function.PostfixMathCommandI#run(java.util.Stack)
	 */
	public void run(Stack inStack) throws ParseException
	{
		// check the stack
		checkStack(inStack);

		// get the parameter from the stack
		Object faces = inStack.pop();
		Object numberOfRolls = inStack.pop();

		// check whether the argument is of the right type
		if (faces instanceof Double && numberOfRolls instanceof Double)
		{
			// calculate the result
			//Integer.MAX_VALUE
			double dRolls = ((Double)numberOfRolls).doubleValue();
			double dFaces = ((Double)faces).doubleValue();
			if (dRolls > Integer.MAX_VALUE || dFaces > Integer.MAX_VALUE)
			{
				throw new ParseException("Values greater than "
						+ Integer.MAX_VALUE
						+ " not allowed.");
			}
			int iRolls = (int)Math.round(((Double)numberOfRolls).doubleValue());
			int iFaces = (int)Math.round(((Double)faces).doubleValue());
			double result = 0;
			result = performRoll(iRolls, iFaces);
			// push the result on the inStack
			inStack.push(new Double(result));
		}
		else
		{
			throw new ParseException("Invalid parameter type");
		}
	}
	
	/**
	 * <p>Performs an actual dice roll.</p>
	 * 
	 * @param iRolls Number of rolls
	 * @param iFaces Number of faces on each die
	 * @return Double result of the roll
	 */
	double performRoll(int iRolls, int iFaces)
	{
		double result = 0;
		RollResult rollLog = null;
		if (iRolls <= 0 || iFaces <= 0)
		{
			if (m_logging)
			{
				if (rollLog == null)
				{
					rollLog = new RollResult(iRolls, iFaces);
				}
				rollLog.add(0);
			}
		}
		else
		{
			for (int i = 0; i < iRolls; i++)
			{
				int roll;
				do
				{
					roll = m_random.choose(0, iFaces) + 1;
				} while (roll > iFaces);
				if (m_logging)
				{
					if (rollLog == null)
					{
						rollLog = new RollResult(iRolls, iFaces);
					}
					rollLog.add(roll);
				}
				result += roll;
			}
		}
		if (m_logging)
		{
			m_rolls.add(rollLog);
		}
		return result;
	}

	/**
	 * <p>Sets the <code>m_random</code> member with a new instance of the requested
	 * <code>RandomElement</code> subclass.  If class loading fails,
	 * initializes <code>m_random</code> member with a <code>DefaultRandom</code>
	 * object.</p>
	 */
	private static void setRandom()
	{
		if (m_random == null || !m_random.getClass().getName().equals(m_randomClassName))
		{
			Class clazz = null;
			try
			{
				clazz = Class.forName(m_randomClassName);
				Class testClass = clazz;
				boolean pass = false;
				while (testClass != null && !pass)
				{
					if ("edu.cornell.lassp.houle.RngPack.RandomElement".equals(testClass.getName()))
					{
						pass = true;
					}
					else
					{
						testClass = testClass.getSuperclass();
					}
				}
				if (pass == true)
				{
					Constructor cons = null;
					try
					{
						cons = clazz.getConstructor(DESIRED_PARAMETER_ARRAY);
						Date[] seed = {new Date()};
						m_random = (RandomElement)cons.newInstance(seed);
					}
					catch (NoSuchMethodException e)
					{
						e.printStackTrace();
						cons = clazz.getConstructor(null);
						m_random = (RandomElement)cons.newInstance(null);
					}
				}
			}
			catch (IllegalAccessException e)
			{
				e.printStackTrace();
			}
			catch (InstantiationException e)
			{
				e.printStackTrace();
			}
			catch (InvocationTargetException e)
			{
				e.printStackTrace();
			}
			catch (NoSuchMethodException e)
			{
				e.printStackTrace();
			}
			catch(SecurityException e)
			{
				e.printStackTrace();
			}
			catch (ClassNotFoundException e)
			{
				e.printStackTrace();
			}
			if (m_random == null)
			{
				m_random = new DefaultRandom();
			}
		}
	}

}
